/*
 * Copyright (C) 2019. Huawei Technologies Co., Ltd. All rights reserved.
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */
#include <stdio.h>
#include <stdlib.h>
#include "irqbalance.h"

extern int keep_going;
extern GMainLoop *main_loop;
extern gboolean scan();
extern int last_interval;
extern int clear_affinity_hint;

int real_sleep_interval;
int sleep_interval_count;
int poll_hint_interval_count;
int sleep_count = 0;
gboolean hint_has_changed = FALSE;

int hint_changed(void)
{
	FILE *file;
	char *line = NULL;
	size_t size = 0;
	gboolean changed = FALSE;

	file = fopen("/proc/irq/affinity_hint_notify", "r+");
	if (!file)
		return changed;

	if (getline(&line, &size, file) > 0 && *line != '0') {
		fprintf(file, "Done");
		changed = TRUE;
	}

	fclose(file);
	if (line)
		free(line);
	return changed;
}

void update_affinity_hint(struct irq_info *info, void *data __attribute__((unused)))
{
	cpumask_t current_affinity_hint;
	char path[PATH_MAX];

	if (!hint_enabled)
		return;

	if (clear_affinity_hint)
		cpus_clear(info->affinity_hint);
	sprintf(path, "/proc/irq/%d/affinity_hint", info->irq);
	process_one_line(path, get_mask_from_bitmap, &current_affinity_hint);

	if (!cpus_equal(current_affinity_hint, info->affinity_hint)) {
		cpumask_copy(info->affinity_hint, current_affinity_hint);
		force_rebalance_irq(info, data);
		hint_has_changed = TRUE;
		cpumask_scnprintf(path, PATH_MAX, current_affinity_hint);
		log(TO_ALL, LOG_INFO, "IRQ(%d): affinity hint modified %s\n", info->irq, path);
	}
}

static void check_clear()
{
	process_one_line("/etc/sysconfig/irqbalance_clear", get_int, &clear_affinity_hint);
}

/*
 * This function is the main loop of irqbalance, which include:
 * 1. scan opration for irq balancing;
 * 2. poll irq affinity hint changes for quickly applying them.
 */
gboolean poll_hint_affinity_and_scan(gpointer data __attribute__((unused)))
{
	gboolean need_verify_flag = FALSE;
	gboolean need_scan_flag = FALSE;

	if (!sleep_interval_count)
		sleep_interval_count = 1;
	if (!poll_hint_interval_count)
		poll_hint_interval_count = 1;

	if (sleep_count % sleep_interval_count == 0) {
		need_scan_flag = TRUE;
	}
	if (sleep_count % poll_hint_interval_count == 0) {
		need_verify_flag = TRUE;
	}
	sleep_count++;

	if (need_verify_flag && hint_changed()) {
		check_clear();
		for_each_irq(NULL, update_affinity_hint, NULL);
		if (hint_has_changed) {
			hint_has_changed = FALSE;
			sleep_count = 1;
			need_scan_flag = TRUE;
		}
	}

	if (need_scan_flag) {
		if (!scan()) {
			g_main_loop_quit(main_loop);
			return FALSE;
		}
	}

	update_interval_and_count();
	if (last_interval != real_sleep_interval) {
		last_interval = real_sleep_interval;
		g_timeout_add_seconds(real_sleep_interval, poll_hint_affinity_and_scan, NULL);
		return FALSE;
	}

	if (keep_going) {
		return TRUE;
	} else {
		g_main_loop_quit(main_loop);
		return FALSE;
	}
}

void update_interval_and_count()
{
	real_sleep_interval =
		sleep_interval > poll_hint_interval ? poll_hint_interval : sleep_interval;
	if (!real_sleep_interval) {
		sleep_interval_count = 1;
		poll_hint_interval_count = 1;
		return;
	}
	sleep_interval_count = sleep_interval / real_sleep_interval;
	poll_hint_interval_count = poll_hint_interval / real_sleep_interval;
}
