/*
 * Copyright (C) 2019. Huawei Technologies Co., Ltd. All rights reserved.
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "irqbalance.h"

char *rules_config_file = NULL;
USER_IRQ_POLICY *user_policy_list = NULL;

void add_usr_irq_policy(USER_IRQ_POLICY *policy)
{
	if (policy == NULL) {
		return;
	}
	policy->next = user_policy_list;
	user_policy_list = policy;
}

char *get_irq_line(int irq)
{
	FILE *file;
	char *line = NULL;
	char *irq_line = NULL;
	size_t size = 0;
	char irq_str[20] = {0}; 
	char *c = NULL;

	snprintf(irq_str, 20, "%d:", irq);

	file = fopen("/proc/interrupts", "r");
	if (getline(&line, &size, file) <= 0) {
		free(line);
		fclose(file);
		return NULL;
	}
	while (getline(&line, &size, file) > 0) {
		c = line;
		while (isblank(*(c)))
			c++;
		if (!isdigit(*c))
			break;

		if (!strncmp(irq_str, c, strnlen(irq_str, 20))) {
			irq_line = line;
			break;
		}
	}

	if (!irq_line)
		free(line);
	fclose(file);
	return irq_line;
}

USER_IRQ_POLICY *get_usr_irq_policy(char *irq_line)
{
	USER_IRQ_POLICY *p = user_policy_list;

	while (p != NULL) {
		if (strstr(irq_line, p->irq_type) != NULL) {
			return p;
		}
		p = p->next;
	}
	return NULL;
}

void set_usr_irq_policy(struct irq_info *info)
{
	USER_IRQ_POLICY *user_policy;
	int irq = info->irq;
	char *irq_line = NULL;

	if (user_policy_list == NULL) {
		return;
	}

	if (irq >= 0)
		irq_line = get_irq_line(irq);
	if (!irq_line)
		return;

	user_policy = get_usr_irq_policy(irq_line);
	if (user_policy != NULL) {
		if (user_policy->numa_node_set) {
			info->numa_node = get_numa_node(user_policy->numa_node);
			log(TO_ALL, LOG_WARNING, "override irq (%d) numa_node to %d\n",
				info->irq, user_policy->numa_node);
		}
		if (user_policy->balance_level_set) {
			info->level = user_policy->balance_level;
			log(TO_ALL, LOG_WARNING, "override irq (%d) balance_level to %d\n",
				info->irq, info->level);
		}
	}

	free(irq_line);
}

int read_user_policy_config()
{
	FILE *file;
	char *line = NULL;
	size_t size = 0;
	size_t len;
	char *key;
	char *value;
	char *c;
	int level;
	int node = -1;
	char savedline[CONFIG_LINE_MAX_LEN] = {0};
	USER_IRQ_POLICY *cur_policy = NULL;
	char *levelvals[] = { "none", "package", "cache", "core" };

	if (rules_config_file == NULL) {
		return 0;
	}
	log(TO_ALL, LOG_INFO, "rules_config_file is: %s\n", rules_config_file);
	file = fopen(rules_config_file, "r");
	if (!file)
		return -1;

	while (!feof(file)) {
		if (getline(&line, &size, file) <= 0)
			break;
		c = line;
		if (*c == '#') {
			continue;
		}
		len = strlen(line);
		if (len > sizeof(savedline)-1) {
			continue;
		}
		strncpy(savedline, line, len);
		savedline[len] = '\0';
		c = savedline;
		while (*c == ' ') {
			c++;
		}
		key = c;
		/* make sure there is no space near '=' */
		c = strchr(savedline, '=');
		if (c != NULL) {
			value = c + 1;
			*c = '\0';
		} else {
			continue;
		}
		c = strchr(value, '\n');
		if (c != NULL) {
			*c = '\0';
		}
		if ((strlen(key) == 0) || (strlen(value) == 0)) {
			continue;
		}
		log(TO_ALL, LOG_INFO, "User irq policy config read: key is %s, value is %s\n", key, value);
		if (strcmp(key, "type") == 0) {
			cur_policy = malloc(sizeof(USER_IRQ_POLICY));
			if (cur_policy == NULL) {
				goto out;
			}
			cur_policy->next = NULL;
			cur_policy->numa_node_set = 0;
			cur_policy->balance_level_set = 0;
			if (strlen(value) > CONFIG_TYPE_MAX_LEN - 1) {
				continue;
			}
			strncpy(cur_policy->irq_type, value, strlen(value) + 1);
			add_usr_irq_policy(cur_policy);
		} else if (strcmp(key, "balance_level") == 0) {
			if (cur_policy == NULL) {
				goto out;
			}
			for (level = 0; level < MAX_LEVEL_NUM; level++) {
				if (strcasecmp(value, levelvals[level]) == 0) {
					break;
				}
			}
			if (level >= MAX_LEVEL_NUM) {
				log(TO_ALL, LOG_WARNING, "Bad value for balance_level policy: %s\n", value);
			} else {
				cur_policy->balance_level = level;
				cur_policy->balance_level_set = 1;
			}
		} else if (strcmp(key, "numa_node") == 0) {
			if (cur_policy == NULL) {
				goto out;
			}
			node = strtoul(value, NULL, 10);
			/* check node */
			if (node != -1 && !get_numa_node(node)) {
				log(TO_ALL, LOG_WARNING, "NUMA node %d doesn't exist\n",
					node);
				continue;
			}
			cur_policy->numa_node = node;
			cur_policy->numa_node_set = 1;
		}
	}
out:
	fclose(file);
	if (line) {
		free(line);
	}
	return 0;
}
